/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 /*
  * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-31
  */

#include <math.h>
#include "T2TAttention.h"
#include "T2TUtility.h"
#include "T2TEmbedding.h"
#include "../../tensor/core/CHeader.h"

namespace transformer
{

/* constructor */
T2TAttention::T2TAttention()
{
    nhead = -1;
    dk = -1;
    dv = -1;
    d = -1;
    isMasked = false;
    ignored = 0;
}

/* deconstructor */
T2TAttention::~T2TAttention()
{
}

/*
initialize the model
>> argc - number of arguments
>> argv - list of pointers to the arguments
>> myIgnored - number of position ignored in attention (from the begining)
>> myIsMasked - indicates whether the attention is with a mask
>> myDevID - device id
*/
void T2TAttention::InitModel(int argc, char** argv,
    bool myIsMasked, int myIgnored,
    int myDevID)
{
    devID = myDevID;
    isMasked = myIsMasked;
    ignored = myIgnored;

    float minmax = 0;

    LoadParamInt(argc, argv, "nhead", &nhead, 4);
    LoadParamInt(argc, argv, "d", &dk, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "d", &dv, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "d", &d, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "maxPosition", &max_relative_position, 8);
    LoadParamFloat(argc, argv, "attminmax", &minmax, 0.1F);
    LoadParamFloat(argc, argv, "dropoutatt", &dropoutP, 0);

    InitTensor2DV2(&wq, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bq, d, X_FLOAT, devID);
    InitTensor2DV2(&wk, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bk, d, X_FLOAT, devID);
    InitTensor2DV2(&wv, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bv, d, X_FLOAT, devID);
    InitTensor2DV2(&rp_embedding_k, max_relative_position * 2 + 1, d/nhead, X_FLOAT, devID);
    InitTensor2DV2(&wo, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bo, d, X_FLOAT, devID);
}

/*
make the network
>> k - keys. It might be of size B * L * H
        where B = batch size, L = sequence length,
        and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
>> cache - layer cache list
>> cacheType - which type that cache is
<< return - multi-attention result
*/
XTensor T2TAttention::Make(XTensor& k, XTensor& q, XTensor& v, XTensor* mask, bool isTraining, Cache* cache, int cacheType)
{
    const bool isEnc = (!cache) ? true : false;
    
    /* linear transformation before self-attention */
    
    XTensor q2, k2, v2;

    q2 = MatrixMul(q, wq) + bq;

    if (!cache) {
        /* self attention for encoder layers */
        k2 = MatrixMul(k, wk) + bk;
        v2 = MatrixMul(v, wv) + bv;
        return MakeRPRAttention(k2, q2, v2, mask, isTraining, isEnc);
    }
    
    else {
        if (cacheType == SELF_ATT) {
            k2 = MatrixMul(k, wk) + bk;
            v2 = MatrixMul(v, wv) + bv;

            /* if hit, we only concat the cache with the new token */
            if (!cache->miss) {
                k2 = Concatenate(cache->key, k2, 1);
                v2 = Concatenate(cache->value, v2, 1);
            }
            cache->key = k2;
            cache->value = v2;
            cache->miss = false;

            return MakeRPRAttention(cache->key, q2, cache->value, mask, isTraining, isEnc);
        }
        else if (cacheType == EN_DE_ATT) {
            if (cache->miss) {
                cache->key = MatrixMul(k, wk) + bk;
                cache->value = MatrixMul(v, wv) + bv;
                cache->miss = false;
            }
            return MakeAttention(cache->key, q2, cache->value, mask, isTraining, isEnc);
        }
        CheckNTErrors(0, "invalid cache type");
    }
}

/*
make the attention network given keys, queries and values (after linear transformation)
>> k - keys. It might be of size B * L * H
        where B = batch size, L = sequence length,
        and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
*/
XTensor T2TAttention::MakeAttention(XTensor& k, XTensor& q, XTensor& v, XTensor* mask, bool isTraining, bool is_encoder)
{
    XTensor kheads;
    XTensor qheads;
    XTensor vheads;

    /* multi head */
    kheads = Split(k, k.order - 1, nhead);
    qheads = Split(q, q.order - 1, nhead);
    vheads = Split(v, v.order - 1, nhead);

    XTensor att;
    XTensor dot;
    XTensor scalar;

    /* scalar = softmax(Q * K^T / sqrt(dk)) * V */
    dot = BMMul(qheads, X_NOTRANS, kheads, X_TRANS);
    
    /*if (isMasked && mask)
        _SumMe(&dot, mask);*/
    
    dot = Linear(dot, 1.0F / (float)sqrt((float)dk / nhead));

    scalar = Softmax(dot, -1);

    if(isTraining && dropoutP > 0)
        scalar = Dropout(scalar, dropoutP);

    att = BMMul(scalar, vheads);

    /* concatenate the heads */
    return MulAndShift(Merge(att, att.order - 1), wo, bo);
}

/*
make the attention network by incorporating the relative position representation with the given keys, queries and values (after linear transformation)
>> k - keys. It might be of size B * L * H
        where B = batch size, L = sequence length,
        and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
*/
XTensor T2TAttention::MakeRPRAttention(XTensor& k, XTensor& q, XTensor& v, XTensor* mask, bool isTraining, bool is_encoder)
{
    XTensor kheads;
    XTensor qheads;
    XTensor vheads;

    const int batch_size = q.GetDim(0);
    const int len_q = q.GetDim(1);
    const int len_kv = k.GetDim(1);

    /* multi head */
    kheads = Split(k, k.order - 1, nhead);
    qheads = Split(q, q.order - 1, nhead);
    vheads = Split(v, v.order - 1, nhead);

    XTensor att;
    XTensor dot;
    XTensor scalar;

    XTensor emb_matrix, relative_key;
    InitTensor2DV2(&emb_matrix, len_q, len_kv, X_INT, q.devID);
    InitTensor3DV2(&relative_key, len_q, len_kv, kheads.GetDim(-1), X_FLOAT, q.devID);
    InitTensor4DV2(&dot, nhead, batch_size, len_q, len_kv, X_FLOAT, q.devID);

    /* generate the relative emb index (L_q, L_kv) */
    GetRPEmbedding(&emb_matrix, len_q, len_kv, max_relative_position, q.devID, is_encoder);

    /* generate the relative key from the rp_embedding_k (L_q, L_kv, H/K) */
    _Gather(&rp_embedding_k, &relative_key, &emb_matrix);

    /* RPR dot product (K, B, L_q, L_kv)*/

    RPDotProduct(&qheads, &kheads, &relative_key, &dot, true);

    /*if (isMasked && mask)
        _SumMe(&dot, mask);*/

    /* scale the dot result */
    dot = Linear(dot, 1.0F / (float)sqrt((float)dk / nhead));

    /* softmax */
    scalar = Softmax(dot, -1);

    if (isTraining && dropoutP > 0)
        scalar = Dropout(scalar, dropoutP);

    /* generate the relative attention output (K, B, L_q, H/K) */
    att = BMMul(scalar, vheads);
    
    /* concatenate the heads */
    return MulAndShift(Merge(att, att.order - 1), wo, bo);
}

void T2TAttention::GetRPEmbedding(XTensor* emb_matrix, const int len_q, const int len_kv, const int max_relative_length, const int devID, const bool is_encoder)
{

    XTensor range;
    InitTensor1DV2(&range, len_kv, X_INT, devID);
    int* index = new int[len_kv];

    // for encoder self-attention which the L_q = L_kv
    if (is_encoder)
    {
        for (int i = 0; i < len_kv; i++)
            index[i] = i;
        range.SetData(index, len_kv);
        XTensor range_2D, range_2D_t;
        InitTensor2DV2(&range_2D, len_q, len_kv, X_INT, devID);
        InitTensor2DV2(&range_2D_t, len_q, len_kv, X_INT, devID);
        _Unsqueeze(&range, &range_2D, 0, len_q);
        _Transpose(&range_2D, &range_2D_t, 0, 1);
        _Sum(&range_2D, &range_2D_t, emb_matrix, -1);
    }
    // for decoder self-attention which the L_q != L_kv, and L_q is 1
    else
    {
        for (int i = 0; i < len_kv; i++)
            index[i] = -len_kv + i + 1;
        range.SetData(index, len_kv);
        _Unsqueeze(&range, emb_matrix, 0, len_q);
    }

    // clip the tensor range from -max_ralative_length to max_relative_length
    _Clip(emb_matrix, emb_matrix, -max_relative_length, max_relative_length);

    // (L_q, L_kv)
    _ScaleAndShift(emb_matrix, emb_matrix, 1, max_relative_length);
    delete[] index;
}

void T2TAttention::RPDotProduct(XTensor* x, XTensor* y, XTensor* z, XTensor* attention, const bool is_key)
{

    const int head_num = nhead;
    const int batch_size = x->dimSize[1];
    const int len_kv = y->dimSize[2];
    const int len_q = x->dimSize[2];
    const int depth = y->dimSize[3];
    // L_kv (is_key=True) or H/K (is_key=False)
    const int last_dim = is_key ? len_kv : depth;
    MATRIX_TRANS_TYPE transpose_flag = is_key ? X_TRANS : X_NOTRANS;

    //if (profiler_) profiler_->StartTimer("RPDotPro-BMM");
    // for key: batch-MM: (K,B,L_q,H/K) * (K,B,H/K,L_kv) -> (K,B,L_q,L_kv)
    // for not key: batch-MM: (K,B,L_q,L_kv) * (K,B,L_kv,H/K) -> (K,B,L_q,H/K)
    XTensor context;
    InitTensor4DV2(&context, head_num, batch_size, len_q, last_dim, X_FLOAT, x->devID);
    _MatrixMulBatched(x, X_NOTRANS, y, transpose_flag, &context);

    // reshape and transpose x to (L_q, K*B, H/K or L_kv)
    int merge_dims[] = { head_num * batch_size, len_q, x->dimSize[3] };
    x->Reshape(3, merge_dims);
    XTensor x_t;
    InitTensor3DV2(&x_t, len_q, head_num * batch_size, x->GetDim(-1), X_FLOAT, x->devID);
    _Transpose(x, &x_t, 0, 1);

    // for key: batch-MM: (L_q, K*B, H/K) * (L_q, L_kv, H/K) -> (L_q, K*B, L_kv)
    // for not key: batch-MM: (L_q, K*B, L_kv) * (L_q, L_kv, H/K) -> (L_q, K*B, H/K)
    XTensor relative;
    InitTensor3DV2(&relative, len_q, head_num * batch_size, last_dim, X_FLOAT, x->devID);
    _MatrixMulBatched(&x_t, X_NOTRANS, z, transpose_flag, &relative);

    // (L_q, K*B, H/K or L_kv) -> (K*B, L_q, H/K or L_kv)
    XTensor relative_t;
    InitTensor3DV2(&relative_t, head_num * batch_size, len_q, last_dim, X_FLOAT, x->devID);
    _Transpose(&relative, &relative_t, 0, 1);

    int split_dims[] = { head_num, batch_size, len_q, last_dim };
    relative_t.Reshape(4, split_dims);

    _Sum(&context, &relative_t, attention);

}
}
