/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University. 
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-08-01
 */

#include <math.h>
#include "T2TEmbedding.h"
#include "T2TUtility.h"
#include "../../tensor/core/CHeader.h"

namespace transformer
{

/* constructor */
T2TEmbedder::T2TEmbedder()
{
    devID = -1;
    vSize = -1;
    maxLength = -1;
}

/* deconstructor */
T2TEmbedder::~T2TEmbedder()
{
}

/* 
initialize the model 
>> argc - number of arguments
>> argv - list of pointers to the arguments
>> myDevID - device id
*/
void T2TEmbedder::InitModel(int argc, char ** argv, int myDevID, bool isEnc)
{
    devID = myDevID;
    
    if(isEnc){
        LoadParamInt(argc, argv, "vsize", &vSize, -1);
    }
    else{
        LoadParamInt(argc, argv, "vsizetgt", &vSize, -1);
    }
    //LoadParamInt(argc, argv, "vsize", &vSize, -1);
    LoadParamInt(argc, argv, "maxlen", &maxLength, 1024);
    LoadParamInt(argc, argv, "d", &eSize, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "d", &d, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "pad", &padIdx, 1);

    InitTensor2DV2(&w, vSize, eSize, X_FLOAT, devID);

    maxLength = maxLength + 1 + 1;
    DTYPE v = 1.0F/(float)sqrt((float)eSize);
    w.SetDataRandn(0, v);

    /* create the positional embedding matrix */
    MakePosEmbedding(eSize, d, maxLength, padIdx);
}

/* 
make positional embeddings (of size eSize * length)
>> eSize - embedding size
>> d - dimension size of the hidden layers
>> length - length of the sequence
*/
void T2TEmbedder::MakePosEmbedding(int eSize, int d, int length, int padIdx)
{
    InitTensor2DV2(&posEmbeddingBase, length, eSize, X_FLOAT, devID);

    float * data = new float[posEmbeddingBase.unitNum];

    for(int pos = 0; pos < length; pos++){
        float * dp = data + pos * eSize;
        
        int channelSize = eSize / 2;
        int offset = 0;
        for(int i = 0; i < channelSize; i++){
            dp[offset++] = (float)sin(pos * exp(-i * log(10000.0F) / (channelSize - 1)));
        }
        for(int i = 0; i < channelSize; i++){
            dp[offset++] = (float)cos(pos * exp(-i * log(10000.0F) / (channelSize - 1)));
        }
        
    }
    
    /* padding zeros */
    int padStart = padIdx * eSize;
    for (int i = padStart; i < padStart + eSize; i++)
        data[i] = 0.F;

    posEmbeddingBase.SetData(data, posEmbeddingBase.unitNum);

    delete[] data;
}

/* 
make the network 
*/
XTensor T2TEmbedder::Make(XTensor &input, int prevLen, int nstep, bool isDec)
{
    
	/* assert padding index is 1 */

	CheckNTErrors(input.order > 1, "Wrong input tensor size!");
	CheckNTErrors(input.dimSize[input.order - 1] < maxLength, "The sequence is too long!");
	CheckNTErrors(vSize > 0, "set vocabulary size by \"-vsize\"");
	CheckNTErrors(eSize > 0, "set embedding size by \"-esize\"");

	XTensor wordEmbedding, position, posEmbedding;
	InitTensorV2(&position, &input);

	int* posData = new int[input.unitNum];

	XTensor inputCPU;
	InitTensorOnCPU(&inputCPU, &input);
	_CopyValues(&input, &inputCPU);


	if (!isDec)
	{
		for (int i = 0; i < inputCPU.GetDim(0); i++) {
			int startNoPad = 2 + prevLen;
			int* p = ((int*)inputCPU.data) + i * inputCPU.GetDim(1);
			for (int j = 0; j < inputCPU.GetDim(1); j++) {
				if (p[j] == 1) {
					posData[i * inputCPU.GetDim(1) + j] = 1;
				}
				else {
					posData[i * inputCPU.GetDim(1) + j] = startNoPad++;
				}
			}
		}
		position.SetData(posData, position.unitNum);
	}
	else
	{
		for (int i = 0; i < position.GetDim(0); i++) {
			for (int j = 0; j < position.GetDim(1); j++) {
				position.Set2DInt(nstep + 2, i, j);
			}
		}

	}


	delete[] posData;

	/* we make positional embeddings first */
	if (true) {
		posEmbedding = Gather(posEmbeddingBase, position);
	}
    /* then we make word embeddings */
    wordEmbedding = Gather(w, input);

    wordEmbedding = Linear(wordEmbedding, (float)sqrt((float)eSize));

    /* we sum over the two embeddings */
	return Sum(wordEmbedding, posEmbedding);
}

}
