/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2019, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 /*
  * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2019-03-27
  */

#include <math.h>
#include "T2TUtility.h"
#include "T2TTester.h"
#include "T2TSearch.h"
#include "../../tensor/XUtility.h"
#include "../../tensor/core/CHeader.h"
#include "../../network/XNoder.h"
#include "..//..//tensor/XTensor.h"

using namespace nts;

namespace transformer
{

/* constructor */
T2TTester::T2TTester()
{
}

/* de-constructor */
T2TTester::~T2TTester()
{
}

/* initialize the model */
void T2TTester::Init(int argc, char** argv)
{
    LoadParamInt(argc, argv, "vsize", &vSize, 34040);
    LoadParamInt(argc, argv, "vsizetgt", &vSizeTgt, vSize);
    LoadParamInt(argc, argv, "sentbatch", &sentBatch, 1);
    LoadParamBool(argc, argv, "sort", &batchLoader.sortBuffer, true);

    seacher.Init(argc, argv);
}

/*
test the model
>> fn - test data file
>> ofn - output data file
>> model - model that is trained
*/
void T2TTester::Test(const char* fn, const char* ofn, T2TModel* model)
{
    int wc = 0;
    int wordCount = 0;
    int wordCountTotal = 0;
    int sentCount = 0;
    int batchCount = 0;

    /* data files */
    FILE* ofile = fopen(ofn, "wb");
    CheckNTErrors(ofile, "Cannot open the output file");

    int devID = model->devID;

    double startT = GetClockSec();

    /* batch of input sequences */
    XTensor batchEnc;

    /* padding */
    XTensor paddingEnc;

    /* an array that keeps the sequences */
    int* seqs = new int[MILLION];

    batchLoader.Init(fn);


    int count = 0;
    while (!batchLoader.IsEmpty())
    {
        count++;
        wordCount = 0;
        for (int i = 0; i < model->decoder->nlayer; ++i) {
            model->decoder->selfAttCache[i].miss = true;
            model->decoder->enDeAttCache[i].miss = true;
        }

        vector<int> indices = batchLoader.LoadBatch(&batchEnc, &paddingEnc, sentBatch, devID);

        XTensor output;
        XTensor score;
        
        seacher.Search(model, &batchEnc, &paddingEnc, &output, &score);
        
        for (int i = 0; i < indices.size(); ++i) {
            Result res;
            XTensor sent, srcIdx, tgtIdx;
            InitTensor1DV2(&srcIdx, 1, X_INT, output.devID);
            int idx[]{ i };
            srcIdx.SetData(idx, 1);
            InitTensorV2(&tgtIdx, &srcIdx);
            SetAscendingOrder(tgtIdx, 0);

            sent = CopyIndexed(output, 0, srcIdx, tgtIdx);
            res.values = sent;
            res.id = indices[i];
            batchLoader.resBuffer.emplace_back(res);
        }

        wc = batchEnc.GetDim(-1);
        wordCount += wc;
        wordCountTotal += wc;
        sentCount += batchEnc.GetDim(-2);
        batchCount += 1;

        if (batchCount % 1 == 0) {
            double elapsed = GetClockSec() - startT;
            XPRINT3(0, stderr, "[INFO] elapsed=%.1fs, sentence=%d, sword=%d\n", elapsed, sentCount, wordCount);
        }
    }

    batchLoader.RerankRes();

    for (auto res : batchLoader.resBuffer) {
        Dump(ofile, &res.values);
    }

    fclose(ofile);

    delete[] seqs;

    double elapsed = GetClockSec() - startT;

    XPRINT3(0, stderr, "[INFO] test finished (took %.1fs, word=%d, sent=%d)\n", elapsed, wordCountTotal, sentCount);
}

/*
dump the result into the file
>> file - data file
>> output - output tensor
*/
void T2TTester::Dump(FILE* file, XTensor* output)
{
    int seqLength = output->GetDim(-1);

    for (int i = 0; i < output->unitNum; i += seqLength) {
        for (int j = 0; j < seqLength; j++) {
            int w = output->GetInt(i + j);
            if (w < 0 || w == 1)
                break;
            fprintf(file, "%d ", w);
            
        }

        fprintf(file, "\n");
    }
}

}
