/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-04-05
*/

#include "DataSet.h"
#include "StringUtil.h"

#include <string>
#include <vector>
#include <fstream>
#include <algorithm>
#include "..//..//..//tensor/XUtility.h"

using namespace nts;

bool Compare(Example& a, Example& b) {
    return a.values.size() > b.values.size();
}

bool CompareRes(Result& a, Result& b) {
    return a.id < b.id;
}

void DataSet::RerankRes(){
    sort(resBuffer.begin(), resBuffer.end(), CompareRes);
}

/*
load data from the file to the buffer
*/
void DataSet::LoadDataToBuffer()
{
    string line;
    buffer.clear();
    bufferUsed = 0;
    const string tokenDelimiter = " ";

    int id = 0;
    while (getline(*fp, line)) {
        vector<int> values = Split<int>(line, tokenDelimiter);

        Example example;
        example.id = id++;
        example.values = values;
        buffer.emplace_back(example);
    }
    if (fp->eof()) {
        fp->seekg(fp->beg);
    }
    if (sortBuffer) {
        sort(buffer.begin(), buffer.end(), Compare);
    }
    resBuffer.reserve(buffer.size());
}

/*
select a field and generate a mini-batch by indices
>>> batchEnc - a tensor to store the batch of input
>>> paddingEnc - a tensor to store the batch of paddings
>>> batchSize - batch size
>>> devID - devices id, -1 for CPU
>>> mem - the memory pool
*/
vector<int> DataSet::LoadBatch(XTensor * batchEnc, XTensor * paddingEnc, 
                        size_t batchSize, int devID)
{
    size_t realBatchSize = batchSize;

    /* real batch size */
    if ((buffer.size()-bufferUsed) < batchSize) {
        realBatchSize = buffer.size()-bufferUsed;
    }

    /* get the maximum sentence length in a mini-batch */
    size_t maxLen = 0;
    if (realBatchSize == 1)
        maxLen = buffer[bufferUsed].values.size();
    for (size_t i = 0; i < realBatchSize - 1; ++i) {
        maxLen = max(maxLen, buffer[bufferUsed+i].values.size());
    }
    CheckNTErrors(maxLen != 0, "wrong length dectected");

    int* batchValues = new int[realBatchSize * maxLen];
    float* paddingValues = new float[realBatchSize * maxLen];

    for (int i = 0; i < realBatchSize * maxLen; ++i) {
        batchValues[i] = 1.0F;
    }
    memset(paddingValues, 0, sizeof(float) * maxLen * realBatchSize);

    size_t cur = 0;
    
    /* left padding */
    vector<int> indices;
    indices.reserve(realBatchSize);
    for (size_t i = 0; i < realBatchSize; ++i) {
        indices.push_back(buffer[bufferUsed + i].id);
        cur = maxLen * (i + 1) - buffer[bufferUsed+i].values.size();
        for (int v : buffer[bufferUsed + i].values) {
            batchValues[cur] = v;
            paddingValues[cur++] = 1.0F;
        }
        cur = maxLen * (i + 1);
    }

    InitTensor2DV2(batchEnc, realBatchSize, maxLen, X_INT, devID);
    InitTensor2DV2(paddingEnc, realBatchSize, maxLen, X_FLOAT, devID);

    bufferUsed += realBatchSize;

    batchEnc->SetData(batchValues, batchEnc->unitNum);
    paddingEnc->SetData(paddingValues, paddingEnc->unitNum);

    delete[] batchValues;
    delete[] paddingValues;

    return indices;
}

/*
the constructor of DataSet
>>> fname - path of the data file
*/
void DataSet::Init(const char* fname)
{
    fp = new ifstream(fname);
    CheckNTErrors(fp->is_open(), "can not open the file");
    bufferUsed = 0;
    
    LoadDataToBuffer();
    if (bufferSize == 0)
        bufferSize = buffer.size();
}

